/*
 * (C) 1995-2004 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */

#define __NO_VERSION

/* PVFS INCLUDES */
#include "config.h"
#include "pvfs_kernel_config.h"
#include "ll_pvfs.h"
#include "pvfsdev.h"
#include "pvfs_v1_xfer.h"

#include "pvfs_linux.h"
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include "kpvfsd.h"

static void exiterror(struct kpvfsd_args *,char *string,int exit_code);
static bool check_for_kpvfsd_stoppage(int kpvfsd_index);
static inline int read_pvfsdev(struct pvfs_upcall *upcall,long timeout);
static inline int write_pvfsdev(struct pvfs_downcall *downcall, long timeout);
static inline void close_pvfsdev(void);
static inline int open_pvfsdev(void);
static inline void init_downcall(struct pvfs_downcall *downcall, struct pvfs_upcall *upcall);
static inline int read_op(struct pvfs_upcall *upcall, struct pvfs_downcall *downcall);
static inline int write_op(struct pvfs_upcall *upcall, struct pvfs_downcall *downcall);

/* KPVFSD related stuff .... */
struct kpvfsd_args *thread_args = NULL;
typedef int (*pfn) (void *);


#if defined(NEW_STYLE_KPVFSD) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
/* This function is executed in the context of keventd a kernel thread in 2.4.xx kernels
 * and should thoretically be a much more lighter way of spawning a kernel thread 
 */
void start_kthread(void *args)
{
	struct kpvfsd_args *targ = (struct kpvfsd_args *)args;
	kernel_thread((pfn)targ->function,targ,0);
	return;
}

pid_t name2pid(char *name)
{
	struct task_struct *p = NULL;
	read_lock(&tasklist_lock);
#ifdef HAVE_FOR_EACH_PROCESS
	for_each_process(p) 
#else
	for_each_task(p) 
#endif
	{
		if(strcmp(p->comm,name) == 0) {
			read_unlock(&tasklist_lock);
			return p->pid;
		}
	}

	read_unlock(&tasklist_lock);
	return -1;
}

#endif

/* this function sets up the kpvfsd kernel threads. There are two ways of doing it,
 * either clone it from the context of the process executing this piece of code 
 * which might be insmod in most cases or spawn it of kevent which is a kernel thread 
 * introduced in kernel 2.4 and onwards.
	returns 0 if successful,
	return -1 if not successful. cleans up after itself. kills all remaining 
					kpvfsd threads, frees up used memory...
	A better way to spawn kernel thread is by calling it in the context of 
	another kernel thread like keventd for example in which case we dont need to do 
	all this stuff of releasing user-space stuff etc ....
	Will do it once this stuff works :-)
*/

int init_kpvfsd(void)
{
	/* 
	* lets try to spawn all the kernel threads, 
	* if any of them fail for whatever reason,
	* we kill of them and free up used memory
	*/
	int j = 0, ret = -ENOMEM, i;
	PDEBUG(D_MOD,"Initializing the kpvfsd threads\n");
	if (num_kpvfsd_threads <= 0) {
		 PDEBUG(D_PVFSD,"Number of kpvfsd threads invalid. Defaulting to 1\n");
		 num_kpvfsd_threads = 1;
	}
	do {
		 thread_args = (struct kpvfsd_args *) kmalloc(num_kpvfsd_threads * sizeof(struct kpvfsd_args), GFP_KERNEL);
		 if (!thread_args) {
				PERROR("Could not allocate memory for semaphores\n");
				break;
		 }
		 for (i = 0; i < num_kpvfsd_threads; i++) {
				init_MUTEX_LOCKED(&thread_args[i].pa_start_stop);
				thread_args[i].pa_id = i;
				thread_args[i].pa_stop = 0;
				thread_args[i].pa_thread = NULL;
				thread_args[i].pa_exit_code = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
				init_waitqueue(&thread_args[i].pa_queue);
#else
				init_waitqueue_head(&thread_args[i].pa_queue);
#endif
				/* ok, let it roll */
#if defined(NEW_STYLE_KPVFSD) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
				thread_args[i].function = kpvfsd;
				INIT_TQUEUE(&thread_args[i].tq,start_kthread,&thread_args[i]);
				/* ask keventd to schedule calling start_kthread at some point of time */
				schedule_task(&thread_args[i].tq);
#else
				/* in older kernels, we spawn off the kernel thread directly */
				kernel_thread((pfn) kpvfsd, (void *) &thread_args[i], 0);
#endif
				down(&thread_args[i].pa_start_stop); 
				if ((ret = thread_args[i].pa_exit_code)) {
					  PERROR("Problem starting kpvfsd thread %d %d\n", i, ret);
					  j = i;
					  break;
				}
		 }
		 /* all threads have been spawned */
		 if (i == num_kpvfsd_threads) {
				return 0;
		 }
	} while (0);
	/* if control reaches here, it is due to some error condition */
	cleanup_kpvfsd(j);
	return ret;
}

/*
 * this function cleans up by killing "num_threads" kpvfsd threads.
 * also frees up used memory.
 * Returns 0 on success and  -1 on failure
 */

int cleanup_kpvfsd(int num_threads)
{
	int i = 0;
	if (num_threads < 0 || num_threads > num_kpvfsd_threads) {
		 PDEBUG(D_PVFSD,"Invalid value of num_threads defaulting:%d\n", num_threads);
		 num_threads = num_kpvfsd_threads;
	}
	/* kill all the spawned kpvfsd threads */
	for (i = 0; i < num_threads; i++) {
		 int err = 0;
		 struct task_struct *thread_instance = NULL;
		 if (!(thread_instance = thread_args[i].pa_thread)) {
				PDEBUG(D_PVFSD,"Killing non existent kpvfsd thread %d \n", i);
				continue;
		 }
		 /* killing must be done with this lock held */
#ifdef NEED_TO_LOCK_KERNEL
		 lock_kernel();
#endif
		 init_MUTEX_LOCKED(&thread_args[i].pa_start_stop);
		 mb();
		 thread_args[i].pa_stop = 1;
		 mb();
		 /* send a SIGKILL to kpvfsd .drop whatever it was doing */
		 if ((err = kill_proc(thread_instance->pid, SIGKILL, 1)) >= 0) {
				/* wait for it to die */
				if(down_interruptible(&thread_args[i].pa_start_stop) != 0) {
						  PERROR("Interrupted! Could not kill kpvfsd %d. Will cause oops!\n",i);
				}
				else {
						  PDEBUG(D_PVFSD,"kpvfsd %d exited with error %d\n",i,thread_args[i].pa_exit_code);
				}
		 }
		 else {
				PERROR("Failure in sending SIGKILL to pid:%d, error:%d\n",thread_instance->pid, err);
		 }
#if defined(NEW_STYLE_KPVFSD) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		 {
			pid_t keventd_pid = 0;
			/* since the thread was spawned by keventd, its status needs to be
			 * * reaped by keventd, else it will be a zombie. Hence we send it a SIGCHLD
			 * * However we need to find the pid of keventd, for that. It is usually 2
			 * * but ....:-)
			 * */
			keventd_pid = name2pid("keventd");
			if((err = kill_proc(keventd_pid,SIGCHLD,1)) < 0) {
				PERROR("Failure in sending SIGCHLD to keventd,%d\n",err);
			}
		 }
#endif
#ifdef NEED_TO_LOCK_KERNEL
		 unlock_kernel();
#endif
	}
	if (thread_args)
		 kfree(thread_args);
	return 0;
}

/*
 * This function returns the index into the pvfsd_args array 
 * which serves as an index for each pvfsd thread.
 * We use current as the key with which we search 
 */

int pvfsd_self(void)
{
	int i = 0;
	for(i=0;i < num_kpvfsd_threads; i++) {
		if(thread_args[i].pa_thread == current) {
			return i;
		}
	}
	return -1;
}

/* the daemonize() function was not there until 2.3 kernels */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
/* Release all user-space context */
static void become_daemon(void)
{
	struct fs_struct *fs = NULL;
	/* release insmod's memory map, use the kernel one's */
	exit_mm(current);
	/* set session leader and process group to init process */
	current->session = 1;
	current->pgrp = 1;
	/* disconnect from the fs usage */
	exit_fs(current);
	fs = init_task.fs;
	current->fs = fs;
	atomic_inc(&fs->count);
	exit_files(current);
	current->files = init_task.files;
	atomic_inc(&current->files->count);
	return;
}
#endif



void thread_exit(struct kpvfsd_args *targs,int exit_code)
{
#ifdef NEED_TO_LOCK_KERNEL
	lock_kernel();// exit will unlock this 
#endif
	targs->pa_thread = NULL;
	targs->pa_exit_code = exit_code;
	up(&targs->pa_start_stop);
	return;
}

/* kpvfsd starts right here too! */
int kpvfsd(struct kpvfsd_args *targs) 
{
	 int err = 0, self = 0;
	 struct pvfs_upcall *upcall = NULL;
	 struct pvfs_downcall *downcall = NULL;
#ifdef NEED_TO_LOCK_KERNEL
	 lock_kernel();
#endif
	 PDEBUG(D_PVFSD,"Starting kpvfsd(pid-%d)\n", current->pid);
	 /*
	  * deamonize and release user-space context , if kpvfsd was launched directly. 
	  * It is already a daemon if launched from keventd 
	  */
#if !defined(NEW_STYLE_KPVFSD)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	 become_daemon();
#else
	 daemonize();
#endif
#endif

#ifdef HAVE_SIGHAND_IN_TASK_STRUCT
	 spin_lock_irq(&current->sighand->siglock);
	 siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGTERM));
	 recalc_sigpending();
	 spin_unlock_irq(&current->sighand->siglock);
#else
	 spin_lock_irq(&current->sigmask_lock);
	 siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGTERM));
	 recalc_sigpending(current);
	 spin_unlock_irq(&current->sigmask_lock);
#endif

	 sprintf(current->comm, "kpvfsd-%d",targs->pa_id);
	 targs->pa_thread = current;
#ifdef NEED_TO_LOCK_KERNEL
	 unlock_kernel();
#endif	 
	 /* Wake up the caller, which is insmod in this case */
	 self = targs->pa_id;
	 if((upcall = (struct pvfs_upcall *)kmalloc(sizeof(struct pvfs_upcall), GFP_KERNEL)) == NULL) {
			exiterror(targs," kpvfsd: Could not allocate memory ofr upcall\n",-ENOMEM);
			return -1; /* will make kpvfsd quit */
	 }
	 if((downcall = (struct pvfs_downcall *)kmalloc(sizeof(struct pvfs_downcall), GFP_KERNEL)) == NULL) {
			kfree(upcall);
			exiterror(targs," kpvfsd: Could not allocate memory for downcall\n",-ENOMEM);
			return -1; /* will make kpvfsd quit */
	 }
	 /*
	  * Currently open_pvfsdev() fails if there is more than 1 kpvfsd daemon.
	  * Need to figure out how to get the master-slave thig working 
	  */
	 if((err = open_pvfsdev()) < 0) {
			kfree(downcall);
			kfree(upcall);
			exiterror(targs, "kpvfsd: Could not open device\n",err);
			return -1; /* will make kpvfsd quit */
	 }
	 if((err = pvfs_comm_init()) < 0) {
			kfree(downcall);
			kfree(upcall);
			exiterror(targs, "kpvfsd: Could not open device\n",err);
			return -1; /* will make kpvfsd quit */
	 }
	 up(&targs->pa_start_stop);
	 /*
		kpvfsd quits only when module is unloaded or the user sends it a signal
		SIGTERM/SIGKILL. However the next time, somebody else does a read/write 
		we should restart this thread since it would not be a functional file system 
		otherwise.
		Basic job is to 
		loop forever, doing:
		 * - read from device
		 * - service request
		 * - write back response
	  */
	 for (;;) {
			memset(upcall,0,sizeof(struct pvfs_upcall));
			err = read_pvfsdev(upcall,30);
			if (err < 0) {
				 PDEBUG(D_PVFSD, "kpvfsd: read_pvfsdev returned error %d \n",err);
				 if(check_for_kpvfsd_stoppage(self) == true) {
						 pvfs_comm_shutdown();
						 close_pvfsdev();
						 kfree(upcall);
						 kfree(downcall);
						 exiterror(targs,"kpvfsd: interrupted/ asked to stop\n",err);
						 return -1; /* will exit kpvfsd */
				 }
			}
			if(err == 0) {
				 PDEBUG(D_PVFSD,"kpvfsd: Idling\n");
				 pvfs_comm_idle();
				 continue;
			}
			/* the do_pvfs_op() call does this already; can probably remove */
			init_downcall(downcall, upcall);
			err = 0;
			switch (upcall->type) {
				 /* all the easy operations */
			case GETMETA_OP:
			case SETMETA_OP:
			case LOOKUP_OP:
			case READLINK_OP:
			case CREATE_OP:
			case REMOVE_OP:
			case RENAME_OP:
			case SYMLINK_OP:
			case MKDIR_OP:
			case RMDIR_OP:
			case STATFS_OP:
			case HINT_OP:
			case FSYNC_OP:
			case LINK_OP:
				 PDEBUG(D_UPCALL, "read upcall; type = %d, name = %s\n", upcall->type, upcall->v1.fhname);
				 /* FALL THROUGH */
			case GETDENTS_OP:
				 err = do_pvfs_op(upcall, downcall);
				 if (err < 0) {
						PDEBUG(D_LIB, "do_pvfs_op failed for type %d\n", upcall->type);
#ifdef REQUEST_DEBUG
						if(upcall->type >= 1 && upcall->type < NUM_OPS) {
								PDEBUG(D_LIB, "do_pvfs_op failed for %s\n",req_str[upcall->type-1]);
						}
#endif
				 }
				 break;
			case READ_OP:
				 err = read_op(upcall, downcall);
				 if (err < 0) {
						PDEBUG(D_LIB, "read_op failed\n");
				 }
				 break;
			case WRITE_OP:
				 err = write_op(upcall, downcall);
				 if (err < 0) {
						PDEBUG(D_LIB, "write_op failed\n");
				 }
				 break;
				 /* things that aren't done yet */
			default:
				 err = -ENOSYS;
				 break;
			}
			/* even if there was an error in any of these operations due to 
			 * a signal, we write a downcall before exiting. 
			 */
			if(err < 0) {
				PDEBUG(D_PVFSD,"Error from the operation %d\n",err);
			}
			downcall->error = err;
			err = 0;
			switch (upcall->type) {
			case HINT_OP:
				 /* this is a one shot hint, we don't want a response */
				 break;
			default:
				 /* the default behavior is to write a response to the device */
				 err = write_pvfsdev(downcall,-1);
				 break;
			}/* end switch */

			/* check if we were interrupted or asked to stop. We ignore the errors from
			 * write_pvfsdev unlike pvfsd.
			 */
			if(err < 0) {
				PDEBUG(D_PVFSD, "Ignoring error from write_pvfsdev : %d\n",err);
			}
			if(check_for_kpvfsd_stoppage(self) == true) {
					  pvfs_comm_shutdown();
					  close_pvfsdev();
					  kfree(upcall);
					  kfree(downcall);
					  exiterror(targs,"kpvfsd: Interrupted by signal/ asked to stop\n",-EINTR);
					  return -1; /* will make kpvfsd quit */
			}
	 }/* end for(;;) */
	 /* shouldnt be coming here */
	 pvfs_comm_shutdown();
	 kfree(upcall);
	 kfree(downcall);
	 exiterror(targs,"kpvfsd: exited outside while(1)\n",-EINVAL);
	 return -1;
}/* end kpvfsd() */


/* read_op()
 *
 * Returns 0 on success, error number on failure.
 * Since kpvfsd resides in the kernel, we dont need to allocate any more
 * memory and can use the memory allocated by the "device" driver.
 */
static inline int read_op(struct pvfs_upcall *upcall, struct pvfs_downcall *downcall)
{
	 int err = 0;
	 if (upcall->u.rw.io.type != IO_CONTIG)
			return -EINVAL;
	 PDEBUG(D_LIB, "read_op: Passed %p %ld\n",upcall->xfer.ptr,(unsigned long)upcall->xfer.size);
	 err = do_pvfs_op(upcall, downcall);
	 if (err < 0) {
			PDEBUG(D_LIB, "read_op: do_pvfs_op failed %d \n",err);
	 }
	 return err;
}

/* write_op()
 *
 * Returns 0 on success, error number on failure.
 * two things can be avoided on using a kernel-space kpvfsd,
 * a) we dont need to allocate any more memory
 * b) we dont need to push the virtual address for the second phase
 *    of the write operation.
 * Hopefully this will translate to big savings!
 */
static inline int write_op(struct pvfs_upcall *upcall, struct pvfs_downcall *downcall)
{
	 int err = 0;
	 if (upcall->u.rw.io.type != IO_CONTIG)
			return -EINVAL;
	 err = do_pvfs_op(upcall, downcall);
	 if (err < 0) {
			PDEBUG(D_LIB, "write_op: do_pvfs_op failed %d\n",err);
	 }
	 return err;
}


static void exiterror(struct kpvfsd_args *targs,char *string,int exit_code)
{
	  if(!targs) {
				 return;
	  }	 
	  PDEBUG(D_PVFSD, "kpvfsd-%d: %s and error code is :%d\n", targs->pa_id,string,exit_code);
	  thread_exit(targs,exit_code);
	  return;
}

/* NOTES ABOUT DEVICE MANIPULATION FUNCTIONS:
	For the old "pvfsd" design, these are the interface functions
	to manipulate the character device interface

	With the new "kpvfsd design, there is really no character device 
	interface, since we can directly deal/operate with the link lists/queues
	built in the device. However as interface functions, the names of most of
	the functions have been retained, which basically means that
	what was earlier a system call would now be a direct function call to
	the functions defined in pvfsdev.c(system call)/kpvfsdev.c(function call)
 */

/* open_pvfsdev()
	kpvfsd: Returns 0 on success, -1 on failure.
*/

static inline int open_pvfsdev(void)
{
	return pvfsdev_open();
}

/* read_pvfsdev()
	kpvfsd: Reads an upcall from the device queues into the pointer specified.
	timeout is currently unused.
	Returns error number on failure and sizeof(upcall) on success
	kpvfsd __blocks__ if there is no upcall to be dequeued.
*/

static inline int read_pvfsdev(struct pvfs_upcall *upcall, long timeout)
{
	return  pvfsdev_read(upcall,sizeof(*upcall),timeout); 
}

/* write_pvfsdev()
	kpvfsd: Writes a downcall into the device queues.
	timeout is currently unused.
	Returns an error number on failure and sizeof(downcall) on success 
*/
static inline int write_pvfsdev(struct pvfs_downcall *downcall, long timeout)
{
	return pvfsdev_write(downcall,sizeof(*downcall),timeout);
}

static inline void close_pvfsdev(void)
{
	pvfsdev_release();
	return;
}

/* init_downcall() - initialize a downcall structure
 */
static inline void init_downcall(struct pvfs_downcall *downcall, struct pvfs_upcall *upcall)
{
	memset(downcall, 0, sizeof(*downcall));
	downcall->magic = PVFS_DOWNCALL_MAGIC;
	downcall->seq = upcall->seq;
	downcall->type = upcall->type;
	return;
}


/* This function is executed in the context of any of the kpvfsd thread to decide 
	if it has to stop or not. 
	There are two reasons for stopping,
	a) Check for bad signals like SIGTERM/ SIGKILL 
	b) unloading of the pvfs-kernel module
	Returns true if pvfsd has to be stopped.
		     false if pvfsd should not be stopped
			  		  or if there any invalid parameters	 
*/
static bool check_for_kpvfsd_stoppage(int pvfsd_index)
{
	struct kpvfsd_args *args = NULL;
	int sig_nr = 0;
	if(pvfsd_index < 0 || pvfsd_index >= num_kpvfsd_threads) {
				return false;
	}
	args = &thread_args[pvfsd_index];
	if(args->pa_stop == 1 || check_for_signal(&sig_nr) < 0) {
			if(args->pa_stop != 1) {
				PDEBUG(D_DEBUG, "kpvfsd got signal %d. will exit shortly....!\n", sig_nr);	  
			}
			return true;
	}
	return false;
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
