/*
 * copyright (c) 1999 Rob Ross and Phil Carns, all rights reserved.
 *
 * Written by Rob Ross and Phil Carns, funded by Scyld Computing.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.clemson.edu
 *           Phil Carns  pcarns@parl.clemson.edu
 */

#include "config.h"
#include "pvfs_kernel_config.h"
#include "pvfs_linux.h"
#ifdef __KPVFSD__
#include "kpvfsd.h"
#include <linux/smp_lock.h>
#else
#include "pvfsd.h"
#endif

/* Ugly hack needed for gcc 3.4.x and broken kernel headers */
#if __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 3)
static const char __module_kernel_version1[] __attribute_used__
       __attribute__((section(".modinfo"))) = "kernel_version=" UTS_RELEASE;
#ifdef CONFIG_MODVERSIONS
static const char __module_using_checksums1[] __attribute_used__ 
		 __attribute__((section(".modinfo"))) = "using_checksums=1";
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9)
static const char __module_license1[] __attribute_used__ 
		 __attribute__((section(".modinfo"))) = "license=GPL";
#endif
#endif


MODULE_AUTHOR("Rob Ross <rbross@parl.clemson.edu> and Phil Carns <pcarns@parl.clemson.edu>");
MODULE_DESCRIPTION("pvfs: PVFS VFS support module");
MODULE_PARM(debug, "1i"); /* one integer debugging parameter */
MODULE_PARM_DESC(debug, "debugging level, 0 is no debugging output");
MODULE_PARM(maxsz, "1i"); /* one integer max size parameter */
MODULE_PARM_DESC(maxsz, "maximum request size, used to limit memory utilization");
MODULE_PARM(buffer, "s"); /* what kind of buffer to use to move data around */
MODULE_PARM_DESC(buffer, "buffering technique, static, dynamic, or mapped");
#ifndef __KPVFSD__
MODULE_PARM(major, "1i");	/* integer major number parameter */
MODULE_PARM_DESC(major, "major number module should attempt to use");
#else
MODULE_PARM(num_kpvfsd_threads, "1i");	/* integer, number of kpvfsd threads to be spawned */
MODULE_PARM_DESC(num_kpvfsd_threads, "Number of kpvfsd kernel threads to be spawned");
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9)
/* New License scheme */
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif


/* we don't need any symbols exported, so let's keep the ksyms table
 * clean.
 */
EXPORT_NO_SYMBOLS;

static int init_pvfs_fs(void);
int pvfsdev_init(void);
void pvfsdev_cleanup(void);

/* GLOBALS */
static int debug = -1;
static int maxsz = -1;
static char *buffer = NULL;
int pvfs_debug = PVFS_DEFAULT_DEBUG_MASK;
int pvfs_buffer = PVFS_DEFAULT_BUFFER;
int pvfs_maxsz = PVFS_DEFAULT_IO_SIZE;

#ifndef __KPVFSD__
static int major = 0;
int pvfs_major = 0;
#else
int num_kpvfsd_threads=1;
#endif

static int __init init_pvfs(void)
{
	int i;
	char b_static[] = "static";
	char b_dynamic[] = "dynamic";
	char b_mapped[] = "mapped";
	char *buftype = NULL;
	PENTRY;

	/* parse module options, print a configuration line */
	if (debug > -1) pvfs_debug = debug;
	if (maxsz >= PVFS_MIN_IO_SIZE) pvfs_maxsz = maxsz;
#ifndef __KPVFSD__
	if (major > 0) pvfs_major = major;
#endif
	if (buffer != NULL) {
		if (strcmp(buffer, b_static) == 0)
			pvfs_buffer = PVFS_BUFFER_STATIC;
#ifdef HAVE_PVFS_BUFMAP
		else if (strcmp(buffer, b_mapped) == 0)
			pvfs_buffer = PVFS_BUFFER_MAPPED;
#endif
		else if (strcmp(buffer, b_dynamic) == 0)
			pvfs_buffer = PVFS_BUFFER_DYNAMIC;
		else {
			printk("buffer value %s is invalid; defaulting to %s\n",
			buffer, b_dynamic);
			pvfs_buffer = PVFS_BUFFER_DYNAMIC;
		}
	}
	switch (pvfs_buffer) {
		case PVFS_BUFFER_DYNAMIC: buftype = b_dynamic; break;
		case PVFS_BUFFER_MAPPED:  buftype = b_mapped;  break;
		case PVFS_BUFFER_STATIC:  buftype = b_static;  break;
		default: printk("pvfs: init_module: should not hit here\n");
	}
	
#ifndef __KPVFSD__
	printk("pvfs: debug = 0x%x, maxsz = %d bytes, buffer = %s, major = %d\n", 
		pvfs_debug, pvfs_maxsz, buftype, pvfs_major);
#else
	/* it looks like the safest option to use is the dynamic option only for kpvfsd */
	buftype = b_dynamic;
	pvfs_buffer = PVFS_BUFFER_DYNAMIC;
	printk ("pvfs: debug = 0x%x, maxsz = %d bytes, buffer = %s\n",pvfs_debug,pvfs_maxsz,buftype);
#endif


	/* initialize all the good stuff */
	i = pvfsdev_init();
	if (i){
		return i;
		PEXIT;
	}
#ifdef __KPVFSD__
	/* Spawn off the kpvfsd kernel thread here */
	i = init_kpvfsd();
	if (i) {
		PERROR("Could not spawn the kpvfsd sub-system: reason :%d\n",i);
		pvfsdev_cleanup();
		return i;
	}
	PDEBUG(D_MOD, "Initialized kpvfsd sub-system\n");
#endif

	i = init_pvfs_fs();
	if(i) {
#ifdef __KPVFSD__
		/* ignore any errors and cleanup the device and kpvfsd layers */
		cleanup_kpvfsd(num_kpvfsd_threads);
#endif
		pvfsdev_cleanup();
	}
	PEXIT;
	return i;
}

static void __exit cleanup_pvfs(void)
{
	PENTRY;
	unregister_filesystem(&pvfs_fs_type);
#ifdef __KPVFSD__
   cleanup_kpvfsd(num_kpvfsd_threads);
#endif
	pvfsdev_cleanup();
	PEXIT;
}

static int init_pvfs_fs(void)
{
	int ret = -1;
	PENTRY;
	ret = register_filesystem(&pvfs_fs_type);
	PEXIT;
	return(ret); 
}

module_init(init_pvfs);
module_exit(cleanup_pvfs);
/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
