/*
 * copyright (c) 1999 Rob Ross and Phil Carns, all rights reserved.
 *
 * Written by Rob Ross and Phil Carns, funded by Scyld Computing.
 *
 * This program is free software; you can redistribute it and/or
 * modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.clemson.edu
 *           Phil Carns  pcarns@parl.clemson.edu
 */

/* 
 * This code could potentially use 3 different methods for memory
 * transfers on read and write depending on which option is defined.
 * PVFS_BUFFER_DYNAMIC - dynamically allocate a buffer for each request
 * PVFS_BUFFER_STATIC - use a single statically allocated buffer
 * PVFS_BUFFER_MAPPED - use the kernel patch to map the user's buffer
 *   into kernel space, avoiding the copy and the allocate altogether.
 */

/* These calls are intended to handle communications between 
 *	the kernel low level vfs calls and the user level pvfsd daemon.
 */

#if defined(__KPVFSD__)
#error "Cannot compile pvfsdev.c with -D__KPVFSD__. Disable this before compiling."
#endif

#include "config.h"

#define __NO_VERSION__
#include "pvfs_kernel_config.h"

#include <linux/poll.h>
#include <linux/iobuf.h>
#include "ll_pvfs.h"
#include "pvfs_bufmap.h"
#include "pvfs_linux.h"
#include "pvfsdev.h"

/* 2.4 */
#ifdef HAVE_LINUX_TQUEUE_H
#include <linux/tqueue.h>
#endif
#include <linux/slab.h>
#include <linux/devfs_fs_kernel.h>
#ifdef HAVE_LINUX_VMALLOC_H
#include <linux/vmalloc.h>
#endif
#include <linux/signal.h>

#include "pvfsd.h"
#include "pvfs_proc.h"
#include "pvfs_mount.h"

/* these operation types are purely for debugging purposes! */
enum {
	TEST_UPWAIT = 667,
	TEST_UPWAIT_COMM = 668,
	TEST_CHEAT_DEC = 669, 
	TEST_RD_XFER = 670, 
	TEST_WR_XFER = 672
};

/* and a maximum sequence number for kernel communications */
/* note: value stolen from INT_MAX macro, but adjusted to reflect
 * the pvfs_seq_t type
 */
#define PVFSDEV_MAX_SEQ ((pvfs_seq_t)(~0U>>1))

extern int pvfs_buffer;
extern int pvfs_maxsz;
extern int pvfs_major;



/********************************************************************
 * file operations:
 */

/* IO operations on the device */
static unsigned int pvfsdev_poll(struct file *my_file, 
											struct poll_table_struct *wait);
static int pvfsdev_ioctl (struct inode * inode, struct file * filp, 
								  unsigned int cmd, unsigned long arg);
static int pvfsdev_open(struct inode * inode, struct file * filp);
static int pvfsdev_release(struct inode * inode, struct file * filp);
static ssize_t pvfsdev_read(struct file *my_file, char *buf, 
								size_t buf_sz, loff_t *offset);
static ssize_t pvfsdev_write(struct file *my_file, const char *buf, 
								 size_t buf_sz, loff_t *offset);

static struct file_operations pvfsdev_fops =
{
	owner:   THIS_MODULE,
	read:    pvfsdev_read,
	write:   pvfsdev_write,
	poll:    pvfsdev_poll,
	ioctl:   pvfsdev_ioctl,
	open:    pvfsdev_open,
	release: pvfsdev_release,
};

/*********************************************************************
 * various structures used in the support code:
 */

/* buffer and size for memory operations */
struct pvfsdev_mem_reg {
	void* ptr;
	pvfs_size_t size;
};

/* this is the internal representation of the upcalls as needed to
 * be placed on a queue
 */
struct pvfsdev_upq_entry {
	struct pvfs_upcall pending_up;
	struct list_head req_list;
};

/* this is the internal representation of the downcalls as needed to
 * be placed on a queue
 */
struct pvfsdev_downq_entry {
	struct pvfs_downcall pending_down;
	struct list_head req_list;
};

/* sequence number entries in the seq_pending queue */
struct pvfsdev_seq_entry {
	pvfs_seq_t seq;
	struct list_head seq_list;
	int invalid;
	int kernel_read; /* flag indicating that data is going to kernel space */

	/* not the most logical place to store this memory transfer
	 * buffer information, but it will keep me from having to
	 * manipulate an extra queue...
	 */

	/* this is the kiobuf for reads and writes to user areas */
	struct kiobuf *iobuf;
	/* this variable is used to determine what phase of the write
	 * operation we are in... may be phased out later.
	 */
	int step;
	/* this is the pointer and offset for reads and writes to kernel
	 * areas
	 */
	struct pvfsdev_mem_reg kernelbuf;
};




/*********************************************************************
 * internal support functions:
 */

/* static int queue_len(struct list_head *my_list); */
static int down_dequeue(struct list_head *my_list, 
	pvfs_seq_t my_seq, struct pvfs_downcall* out_downcall); 
static int seq_is_pending(struct list_head *my_list, pvfs_seq_t
	my_seq, int* invalid);
static int pvfsdev_down_enqueue(struct pvfs_downcall *in_downcall);
static int remove_seq_pending(struct list_head *my_list, pvfs_seq_t my_seq);
static int invalidate_seq_pending(struct list_head *my_list, pvfs_seq_t 
	my_seq);
static pvfs_seq_t inc_locked_seq(void);
static int up_dequeue(struct list_head *my_list, pvfs_seq_t my_seq);
static struct pvfsdev_seq_entry* match_seq_entry(struct list_head *my_list, 
	pvfs_seq_t my_seq);
static int is_fake_useraddr(void);
static inline int pvfsdev_read_in_data(void *kptr, char *uptr, 
													pvfs_size_t size);
static inline int pvfsdev_write_out_data(char *uptr, void *kptr, 
													  pvfs_size_t size);
static inline int check_for_bad_signals(int *sig_nr);
static inline void cleanup_buffer(struct pvfs_upcall *in_upcall,
											 void **copybufp, void **pvfs_map_bufp);
static int setup_buffer(struct pvfs_upcall *in_upcall, 
								void **copybufp, void **pvfs_map_bufp);
/**********************************************************************
 * various device internal queues and other device wide structures
 * that require locking before use. 
 */

/* this will be the queue of pending messages to pvfsd */
LIST_HEAD(up_pending);

/* this is the queue of upcall sequence numbers to be matched 
 * with the corresponding downcall in pvfsdev_enqueue.
 */
LIST_HEAD(seq_pending);

/* this is the list of pending downcalls */
LIST_HEAD(down_pending);

/* this the sequence number assigned to upcalls */
static pvfs_seq_t current_seq = 1;

/* this variable keeps up with the number of processes which have
 * opened the device at any given time.  Mostly for sanity checking
 * so that pvfs ops aren't attempted with no pvfsd.
 */
static int nr_daemons;

/* this is a buffer which will be used in transfers between the
 * application and pvfsd if static copies are used (the "static" buffer
 * option)
 */
static void *static_buf = NULL;

/* two semaphores:
 * static_buf_sem - protects our static buffer
 * seq_sem - protects sequence list
 */
static DECLARE_MUTEX(static_buf_sem);
static DECLARE_MUTEX(seq_sem);

/* keeps up with the amount of memory allocated for dynamic 
 * and mapped buffers 
 */
static unsigned long pvfsdev_buffer_size = 0;

/* these locks are used to protect the data defined above */
static spinlock_t pvfsdev_up_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_down_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_current_seq_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_nr_daemons_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_buffer_pool_lock = SPIN_LOCK_UNLOCKED;

/* three device wait queues:
 * pvfsdev_waitq - used for polling
 * pvfsdev_seq_wait - used for sleeping while waiting on a
 *   sequence-matched downcall
 * pvfsdev_buffer_pool_wait - used to sleep when dynamic buffer has
 *   gotten too large
 */
static DECLARE_WAIT_QUEUE_HEAD(pvfsdev_waitq);
static DECLARE_WAIT_QUEUE_HEAD(pvfsdev_seq_wait);
static DECLARE_WAIT_QUEUE_HEAD(pvfsdev_buffer_pool_wait);

#ifdef HAVE_DEVFS
/* device handle for the character device */
static devfs_handle_t devfs_handle = NULL;
#endif

/**********************************************************************
 * module specific code :
 */

/* pvfsdev_init()
 *
 * Called at module load time.
 */
int pvfsdev_init(void)
{
	int ret;       /* Dynamically assigned major number */
	/* do I need to do more stuff here?  how 'bout checking the minor
	 * number? 
	 */

#ifdef HAVE_DEVFS
	if ((ret = devfs_register_chrdev(pvfs_major, "pvfsd", &pvfsdev_fops)) < 0)
#else
	if ((ret = register_chrdev(pvfs_major, "pvfsd", &pvfsdev_fops)) < 0)
#endif
	{
		/* failed to register device */
		if(pvfs_major){
			PERROR("pvfsdev : unable to get major %d\n", PVFSD_MAJOR);
		}
		else{
			PERROR("pvfsdev : could not dynamically assign major number.\n");
		}

		return -ENODEV;
	}
	else{
		if(ret > 0){
			/* major number was dynamically assigned */
			pvfs_major = ret;
		}
	}

#ifdef HAVE_DEVFS
	devfs_handle = devfs_register(NULL, "pvfsd", DEVFS_FL_DEFAULT,
	pvfs_major, 0, (S_IFCHR | S_IRUSR | S_IWUSR), &pvfsdev_fops, NULL);
#endif

	/* initialize the sequence counter */
	spin_lock(&pvfsdev_current_seq_lock);
	current_seq = 0;
	spin_unlock(&pvfsdev_current_seq_lock);
	spin_lock(&pvfsdev_nr_daemons_lock);
	nr_daemons = 0;
	spin_unlock(&pvfsdev_nr_daemons_lock);

	if (pvfs_buffer == PVFS_BUFFER_STATIC) {
		down(&static_buf_sem);
		static_buf = vmalloc(pvfs_maxsz);
		up(&static_buf_sem);
		if (static_buf == NULL) return -ENOMEM;
	}

	PDEBUG(D_PSDEV, "pvfsdev: device module loaded (major=%d)\n", pvfs_major);
	pvfs_sysctl_init();
	return 0;
}

/* pvfsdev_cleanup()
 *
 * Called at module unload.
 */
void pvfsdev_cleanup(void)
{
	struct pvfsdev_upq_entry *up_check;
	struct pvfsdev_downq_entry *down_check;
	struct pvfsdev_seq_entry *seq_check;
	int rm_seqs = 0;
	int inv_seqs = 0;
   pvfs_sysctl_clean();

	/* we need to free up the all of the pending lists at this point */

	spin_lock(&pvfsdev_up_lock);
	while (up_pending.next != &up_pending) {
		up_check = list_entry(up_pending.next, struct pvfsdev_upq_entry, 
									 req_list);
		list_del(up_pending.next);
		kfree(up_check);
	}
	spin_unlock(&pvfsdev_up_lock);

	spin_lock(&pvfsdev_down_lock);
	while (down_pending.next != &down_pending) {
		down_check = list_entry(down_pending.next, struct pvfsdev_downq_entry, 
										req_list);
		list_del(down_pending.next);
		kfree(down_check);
	}
	spin_unlock(&pvfsdev_down_lock);

	down(&seq_sem);
	while (seq_pending.next != &seq_pending) {
		seq_check = list_entry(seq_pending.next, struct pvfsdev_seq_entry, 
									  seq_list);
		rm_seqs++;
		if(seq_check->invalid){
			inv_seqs++;
		}
		list_del(seq_pending.next);
		kfree(seq_check);
	}
	up(&seq_sem);

	if(rm_seqs){
		PERROR("pvfsdev: %d upcall sequences removed, %d of which had become invalid.\n", rm_seqs, inv_seqs);
	}

	if (pvfs_buffer == PVFS_BUFFER_STATIC) {
		/* get rid of the static buffer we had laying around */
		down(&static_buf_sem);
		if (static_buf != NULL) {
			vfree(static_buf);
			static_buf = NULL;
		}
		up(&static_buf_sem);
	}

#ifdef HAVE_DEVFS
	devfs_unregister(devfs_handle);
	devfs_unregister_chrdev(pvfs_major, "pvfsd");
#else
	unregister_chrdev(pvfs_major, "pvfsd");
#endif
	PDEBUG(D_PSDEV, "pvfsdev: device module unloaded\n");
	return;
}


/***************************************************************
 * file operation code:
 */

/* pvfsdev_open()
 *
 * Called when a process opens the PVFS device
 *
 * Returns 0 on success, -errno on failure.
 */
static int pvfsdev_open(struct inode *inode, struct file *filp)
{

	PDEBUG(D_PSDEV, "pvfsdev: device opened.\n");
	
	/* we only allow one device to exist on the system */
	if ((MINOR(inode->i_rdev)) != 0) {
		return(-ENODEV);
	}

	MOD_INC_USE_COUNT; /* moved here in 2.4 code */

	spin_lock(&pvfsdev_nr_daemons_lock);
	/* we really only allow one daemon to use this device at a time */
	if (nr_daemons != 0) {
		spin_unlock(&pvfsdev_nr_daemons_lock);
		MOD_DEC_USE_COUNT; /* moved here in 2.4 code */
		return(-EBUSY);
	}
	nr_daemons++;
	spin_unlock(&pvfsdev_nr_daemons_lock);

	return(0);
}

/* pvfsdev_release()
 *
 * Called when process closes the device
 */
static int pvfsdev_release(struct inode * inode, struct file * filp)
{

	PDEBUG(D_PSDEV, "pvfsdev: device released.\n");

	spin_lock(&pvfsdev_nr_daemons_lock);
	nr_daemons--;
	spin_unlock(&pvfsdev_nr_daemons_lock);
	
	MOD_DEC_USE_COUNT;

	return(0);
}


/* pvfsdev_write()
 *
 * Called when the PVFS daemon writes into the PVFS device file.
 *
 * Returns 0 on success, -errno on failure.
 */
static ssize_t pvfsdev_write(struct file *my_file, const char *buf, 
	size_t buf_sz, loff_t *offset){

	struct pvfs_downcall in_req;
	int req_sz = sizeof(struct pvfs_downcall);
	int ret = 0;
	int invalid = 0;
	struct pvfsdev_seq_entry *my_seq = NULL;

	PDEBUG(D_DOWNCALL, "pvfsdev: somebody is writing to me\n");
	if(buf_sz != req_sz){
		return -EINVAL;
	}
	
	/* grab the request from user space so that we can manipulate it */
	ret = copy_from_user(&in_req, buf, buf_sz);
	if (ret != 0) {
		PERROR("pvfsdev: error from copy_from_user.\n");
		return -ENOMEM;
	}

	/* verify the magic */
	if (in_req.magic != PVFS_DOWNCALL_MAGIC) {
		PERROR("pvfsdev: bad downcall magic.\n");
		return -EINVAL;
	}

	if (in_req.seq == -1) {
		/* this is an independent downcall, do not match it to
		 * an outstanding upcall sequence. 
       */
		/* do something else instead- not defined yet. */
		PERROR("pvfsdev: unexpected independent downcall.\n");
		return -ENOSYS;
	}

	PDEBUG(D_TIMING, "pvfsdev: writing downcall number %d.\n",
		in_req.seq);

	/* ok- we are going to grab this semaphore right here, and then hang
	 * on to it for a long time to make sure that no one invalidates our
	 * sequence number while we are completing this downcall!
	 */
	down(&seq_sem);

	/* check to see if we are expecting this sequence number */
	ret = seq_is_pending(&seq_pending, in_req.seq, &invalid);
	if(!ret){
		/* Uh-oh... something is amok because we don't have the
		 * sequence number for this downcall queued up... bail
		 * out! 
       */
		up(&seq_sem);
		PDEBUG(D_TIMING, "pvfsdev: can't match sequence number.\n");
		return -EINVAL;
	}

	/* at this point we need to be sure this write is still valid
	 * with respect to its comm sequence.  If not just eat it and
	 * return as normal- the user app on the the other side was
	 * probably terminated. 
    */
	if (invalid) {
		/* what sort of op has been invalidated? */
		my_seq = match_seq_entry(&seq_pending, in_req.seq);
		if(in_req.type == WRITE_OP && !my_seq->step) {
			/* this is the _first_ step in a write operation */
			/* we can't remove this sequence until the 2nd step */
			PDEBUG(D_TIMING, "pvfsdev: invalid seq in 1st write step.\n");
			my_seq->step = 1;
			/* do nothing for now; cleanup on next step */
			up(&seq_sem);
			return req_sz;
		}

		PDEBUG(D_TIMING, "pvfsdev: recieved write for invalid seq.\n");
		ret = remove_seq_pending(&seq_pending, in_req.seq);
		up(&seq_sem);
		if(ret < 0){
			PDEBUG(D_TIMING, "pvfsdev: remove of seq_entry failed... \n");
		}
		return req_sz;
	}
	 
	/* now everything should be Ok for submitting this downcall */

	/* use the switch here to catch special test cases as well as
	 * normal requests that require memory transfers. 
    */

	switch (in_req.type) {
		case WRITE_OP:
			/* There are two separate downcalls for writes.  The first one is
			 * used by the pvfsd to indicate the location in its address space
			 * where data should be placed.  The second indicates the result of
			 * the PVFS request it performed on behalf of the application
			 * process.  
          */
			PDEBUG(D_SPECIAL, "pvfsdev: performing memory xfer (write).\n");

			if ((my_seq = match_seq_entry(&seq_pending, in_req.seq)) == NULL) {
				up(&seq_sem);
				PDEBUG(D_SPECIAL, "pvfsdev: couldn't find iobuf!\n");
				return -EINVAL;
			}

			/* Here is where we determine if this is the first or second downcall
          * for this write operation.  The my_seq->step value is 0 prior to
          * the first downcall.  We set it to 1 to indicate on the next pass
			 * that we have already received the first downcall for this 
			 * operation.
          */
			if (!my_seq->step) {
				my_seq->step = 1;
								
				PDEBUG(D_SPECIAL, "pvfsdev: preparing to copy %ld bytes.\n", 
						 (long)in_req.xfer.size);
				if (my_seq->kernel_read) {
					/* pulling data from kernel space */
					ret = pvfsdev_write_out_data(in_req.xfer.ptr, 
														  my_seq->kernelbuf.ptr, 
														  in_req.xfer.size);
				}
				else {
					ret = pvfsdev_write_out_data(in_req.xfer.ptr,
														  my_seq->iobuf, 
														  in_req.xfer.size);
				}

				if (ret != 0) {
					up(&seq_sem);
					PDEBUG(D_SPECIAL, "pvfsdev: pvfs memory transfer failure.\n");
					return -ENOMEM;
				}
		
				/* we don't enqueue the first of the two write downcalls; 
             * just return instead.
				 */
				up(&seq_sem);
				return req_sz;
			}
			
			break;
		case TEST_RD_XFER:
		case READ_OP:
			PDEBUG(D_SPECIAL, "pvfsdev: performing memory xfer (read).\n");
			if((my_seq = match_seq_entry(&seq_pending, in_req.seq)) == NULL){
				PERROR("pvfsdev: couldn't find iobuf!\n");
				up(&seq_sem);
				return -EINVAL;
			}
			PDEBUG(D_SPECIAL, "pvfsdev: preparing to copy %ld bytes.\n",
					 (long)in_req.xfer.size);
			if (my_seq->kernel_read) {
				/* storing data in kernel space */
				ret = copy_from_user(my_seq->kernelbuf.ptr, 
					in_req.xfer.ptr, 
					in_req.xfer.size);
			}
			else {
				ret = pvfsdev_read_in_data(my_seq->iobuf, 
													in_req.xfer.ptr, 
													in_req.xfer.size);
			}

			if (ret != 0) {
				up(&seq_sem);
				PDEBUG(D_SPECIAL,
				"pvfsdev: copy_from_user/pvfs_copy_to_userbuf failure.\n");
				return -ENOMEM;
			}
			break;
		case GETDENTS_OP:
			PDEBUG(D_SPECIAL, "pvfsdev: performing memory xfer (getdents).\n");
			if ((my_seq = match_seq_entry(&seq_pending, in_req.seq)) == NULL) {
				up(&seq_sem);
				PDEBUG(D_SPECIAL, "pvfsdev: couldn't find kernelbuf!\n");
				return -EINVAL;
			}

			ret = copy_from_user(my_seq->kernelbuf.ptr, in_req.xfer.ptr, 
				in_req.xfer.size);
			if (ret != 0) {
				up(&seq_sem);
				PDEBUG(D_SPECIAL, 
						 "pvfsdev: copy_from_user to pvfs_userbuf failure.\n");
				return -ENOMEM;
			}
			break;
		case READLINK_OP:
			PDEBUG(D_SPECIAL, "pvfsdev: performing memory xfer (readlink).\n");
			if ((my_seq = match_seq_entry(&seq_pending, in_req.seq)) == NULL) {
				up(&seq_sem);
				PDEBUG(D_SPECIAL, "pvfsdev: couldn't find kernelbuf!\n");
				return -EINVAL;
			}
			PDEBUG(D_SPECIAL, "pvfsdev: copying %ld bytes from userspace %p to kernelspace %p\n", (long)in_req.xfer.size, my_seq->kernelbuf.ptr, in_req.xfer.ptr);

			ret = copy_from_user(my_seq->kernelbuf.ptr, in_req.xfer.ptr, 
				in_req.xfer.size + 1);
			if (ret != 0) {
				up(&seq_sem);
				PDEBUG(D_SPECIAL, 
						 "pvfsdev: copy_from_user to pvfs_userbuf failure.\n");
				return -ENOMEM;
			}
			break;
		default:
			/* no memory transfer required for this type of request. */
			break;
	}
	
	ret = pvfsdev_down_enqueue(&in_req);
	if (ret < 0) {
		up(&seq_sem);
		return ret;
	}
	
	up(&seq_sem);
	return req_sz;
}


/* pvfsdev_read_in_data()
 *
 * Parameters:
 * kptr - kernel space destination, could be an address or a kiobuf
 * uptr - user space (source) address pointer
 * size - size of data to move
 */
static inline int pvfsdev_read_in_data(void *kptr, char *uptr, 
													pvfs_size_t size)
{
	int ret;
#ifdef HAVE_PVFS_BUFMAP
	if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
      /* in this case we are copying from one userspace to another */
		ret = pvfs_copy_to_userbuf((struct kiobuf *)kptr, uptr, size);
		return ret;
	}
#endif
   ret = copy_from_user((char *)kptr, uptr, size);
   return ret;
}

/* pvfsdev_write_out_data()
 *
 * Parameters:
 * uptr - user space (destination) address pointer
 * kptr - kernel space source, could be an address or a kiobuf
 * size - size of data to move
 */
static inline int pvfsdev_write_out_data(char *uptr, void *kptr, 
													  pvfs_size_t size)
{
	int ret;
#ifdef HAVE_PVFS_BUFMAP
	if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
      /* in this case we are copying from one userspace to another */
		ret = pvfs_copy_from_userbuf(uptr, (struct kiobuf *)kptr, size);
		return ret;
	}
#endif
   ret = copy_to_user(uptr, (char *)kptr, size);
   return ret;
}



/* pvfsdev_read()
 *
 * Called when pvfsd reads from the PVFS device.
 *
 * Returns 0 on success, -errno on failure.
 */
static ssize_t pvfsdev_read(struct file *my_file, char *buf, size_t buf_sz,
loff_t *offset)
{

	struct pvfsdev_upq_entry *out_entry;
	int retval = 0;
	DECLARE_WAITQUEUE(wait, current);

	PDEBUG(D_UPCALL, "pvfsdev: someone is reading from me.\n");
	
	if (buf_sz != sizeof(struct pvfs_upcall)) {
		return -EINVAL;
	}

	/* put ourselves on a wait queue while we check to see if there is
	 * really an upcall available to read.  If it isn't there, we will
	 * schedule ourselves to wait on it.
	 */
	wait.task = current;
	add_wait_queue(&pvfsdev_waitq, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	while(up_pending.prev == &up_pending){
		if(my_file->f_flags & O_NONBLOCK){
			retval = -EAGAIN;
			break;
		}
		if(signal_pending(current)){
			retval = -ERESTARTSYS;
			break;
		}
		schedule();
	}

	remove_wait_queue(&pvfsdev_waitq, &wait);
	set_current_state(TASK_RUNNING);

	if(retval){
		/* we were interrupted or are in non-blocking mode w/ nothing to do */
		return(retval);
	}

	/* if we reach this point in the code, there should be an upcall to
	 * read, unless the other side somehow manages to abort before we
	 * actually read it out
	 */

	/* get the upcall off of the queue and copy it to the user (pvfsd) */
	spin_lock(&pvfsdev_up_lock);
	if (up_pending.prev != &up_pending) {
		out_entry = list_entry(up_pending.prev, struct pvfsdev_upq_entry, 
									  req_list);
		PDEBUG(D_TIMING, "pvfsdev: reading upcall number %d.\n",
			out_entry->pending_up.seq);
		list_del(up_pending.prev);
		spin_unlock(&pvfsdev_up_lock);
		copy_to_user(buf, &(out_entry->pending_up), sizeof(struct pvfs_upcall));
		kfree(out_entry);
		return(sizeof(struct pvfs_upcall));
	}
	else {
		/* nothing in the queue */
		spin_unlock(&pvfsdev_up_lock);
		return 0;
	}
}

/* poll (was select in 2.0 kernel) */
static unsigned int pvfsdev_poll(struct file *my_file, struct 
	poll_table_struct *wait){

	/* do I need to increment the module count here? */
	/* nah, that would probably be bad if process selecting on device
	 * prevents it from unloading... */

	unsigned int mask = POLLOUT | POLLWRNORM;
	poll_wait(my_file, &pvfsdev_waitq, wait);
	if(!(list_empty(&up_pending))){
		mask |= POLLIN | POLLRDNORM;
	}
		
	return mask;
}

/* ioctl is mostly used for debugging right now. */
static int pvfsdev_ioctl (struct inode * inode, struct file * filp, 
	unsigned int cmd, unsigned long arg){

	struct pvfs_upcall *loop_up = NULL;
	struct pvfs_downcall *test_req = NULL;
	int ret = 0;
	
	struct ioctl_mem_reg {
		void* ptr;
		pvfs_size_t size;
	} rd_xfer_reg;

	PDEBUG(D_PIOCTL, "pvfsdev: ioctl.\n");

	switch(cmd){
		case(TEST_UPWAIT):
			/* this will exercise the communication mechanism that
			 * sends an upcall and sleeps until the matching downcall
			 * arrives - for testing purposes during development
			 */
			 
			PDEBUG(D_PIOCTL, "pvfsdev: initialized test_upwait\n");

			loop_up = kmalloc(sizeof(struct pvfs_upcall), GFP_KERNEL);
			if(!loop_up){
				kfree(loop_up);
				return(-ENOMEM);
			}
			loop_up->magic = PVFS_UPCALL_MAGIC;
			loop_up->type = TEST_UPWAIT_COMM;

			test_req = kmalloc(sizeof(struct pvfs_downcall), GFP_KERNEL);
			if(!test_req){
				kfree(test_req);
				kfree(loop_up);
				return(-ENOMEM);
			}
			ret = pvfsdev_enqueue(loop_up, test_req);
			if(ret < 0){
				kfree(test_req);
				kfree(loop_up);
				return(ret);
			}
			kfree(test_req);
			kfree(loop_up);
			return(0);
		case(TEST_RD_XFER):
			/* this will exercise the communication mechanism that
			 * can transfer a variable length memory region from one
			 * user program to another.
			 */
			 
			PDEBUG(D_PIOCTL, "pvfsdev: initialized test_rd_xfer\n");
			copy_from_user(&rd_xfer_reg, (struct ioctl_mem_reg*)arg, sizeof(struct
				ioctl_mem_reg));

			loop_up = kmalloc(sizeof(struct pvfs_upcall), GFP_KERNEL);
			if(!loop_up){
				kfree(loop_up);
				return(-ENOMEM);
			}

			loop_up->magic = PVFS_UPCALL_MAGIC;
			loop_up->type = TEST_RD_XFER;
			loop_up->xfer.ptr = rd_xfer_reg.ptr;
			loop_up->xfer.size = rd_xfer_reg.size;

			test_req = kmalloc(sizeof(struct pvfs_downcall), GFP_KERNEL);
			if(!test_req){
				kfree(test_req);
				kfree(loop_up);
				return(-ENOMEM);
			}
			ret = pvfsdev_enqueue(loop_up, test_req);
			if(ret < 0){
				kfree(test_req);
				kfree(loop_up);
				return(ret);
			}
			kfree(test_req);
			kfree(loop_up);
			return(0);

		case(TEST_WR_XFER):
			/* this will exercise the communication mechanism that
			 * can transfer a variable length memory region from one
			 * user program to another.
			 */
			 
			PDEBUG(D_PIOCTL, "pvfsdev: initialized test_wr_xfer\n");
			copy_from_user(&rd_xfer_reg, (struct ioctl_mem_reg*)arg, sizeof(struct
				ioctl_mem_reg));

			loop_up = kmalloc(sizeof(struct pvfs_upcall), GFP_KERNEL);
			if(!loop_up){
				kfree(loop_up);
				return -ENOMEM;
			}

			loop_up->magic = PVFS_UPCALL_MAGIC;
			loop_up->type = TEST_WR_XFER;
			loop_up->xfer.ptr = rd_xfer_reg.ptr;
			loop_up->xfer.size = rd_xfer_reg.size;

			test_req = kmalloc(sizeof(struct pvfs_downcall), GFP_KERNEL);
			if(!test_req){
				kfree(test_req);
				kfree(loop_up);
				return -ENOMEM;
			}
			ret = pvfsdev_enqueue(loop_up, test_req);
			if(ret < 0){
				kfree(test_req);
				kfree(loop_up);
				return ret;
			}
			kfree(test_req);
			kfree(loop_up);
			return 0;


		case(TEST_CHEAT_DEC):
			/* this is for development purposes...
			 * manually decrements the module use count */
			PDEBUG(D_PIOCTL, "pvfsdev: manually decrementing module count.\n");
			MOD_DEC_USE_COUNT;
			return 0;
		default:
			return -ENOSYS;
	}
	
}




/*******************************************************************
 * support functions:
 */


#if 0
/* queue_len()
 *
 * for debugging, (and also pvfsdev_poll) see how long the outgoing
 * queue is..
 *
 * As of 7-14-2001 this is no longer being used anywhere.  We are just
 * leaving it in here in case it is needed for debugging purposes.
 */
static int queue_len(struct list_head *my_list){
	struct list_head *check;
	int count = 0;

	spin_lock(&pvfsdev_up_lock);
	check = my_list->next;
	while(check != my_list){
		count++;
		check = check->next;
	}
	spin_unlock(&pvfsdev_up_lock);
	return count;
}
#endif

/*
 *	Export the signal mask handling for code that
 *	may sleep for the operation to complete.
 *	Only if intr is 0, do we change the process's
 *	signal dispositions, 
 *	save the old dispostion in oldset and return.
 *	Otherwise we don't do anything.
 */
 
static void sig_setmask(sigset_t *oldset, int intr)
{
	/* Always allow SIGKILL */
	unsigned long	sigallow = sigmask(SIGKILL);
	unsigned long	irqflags;
	struct k_sigaction *action;

	/* If it was intr-mounted, then this function is a nop */
	if(intr) {
		return;
	}
#ifdef HAVE_SIGHAND_IN_TASK_STRUCT
	action = current->sighand->action;
	if (action[SIGINT-1].sa.sa_handler == SIG_DFL)
		sigallow |= sigmask(SIGINT);
	if (action[SIGQUIT-1].sa.sa_handler == SIG_DFL)
		sigallow |= sigmask(SIGQUIT);
	spin_lock_irqsave(&current->sighand->siglock, irqflags);
	*oldset = current->blocked;
	siginitsetinv(&current->blocked, sigallow & ~oldset->sig[0]);
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, irqflags);
#else
	action = current->sig->action;
	if (action[SIGINT-1].sa.sa_handler == SIG_DFL)
		sigallow |= sigmask(SIGINT);
	if (action[SIGQUIT-1].sa.sa_handler == SIG_DFL)
		sigallow |= sigmask(SIGQUIT);
	spin_lock_irqsave(&current->sigmask_lock, irqflags);
	*oldset = current->blocked;
	siginitsetinv(&current->blocked, sigallow & ~oldset->sig[0]);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, irqflags);
#endif
	return;
}

/* restore the old signal handling disposition only if intr was not set */

static void sig_setunmask(sigset_t *oldset, int intr)
{
	unsigned long	irqflags;

	/* If it was intr-mounted, then this function is a nop */
	if(intr) {
		return;
	}
#ifdef HAVE_SIGHAND_IN_TASK_STRUCT
	spin_lock_irqsave(&current->sighand->siglock, irqflags);
	current->blocked = *oldset;
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, irqflags);
#else 
	spin_lock_irqsave(&current->sigmask_lock, irqflags);
	current->blocked = *oldset;
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, irqflags);
#endif
	return;
}

/* pvfsdev_enqueue()
 *
 * this adds upcalls to the outgoing queue.  it will then wait on a matching
 * downcall before returning.
 *
 * TODO: Optimize better for the no-patch case when we feel like it.
 * Returns 0 on success, -errno on failure.
 *
 * NOTE: We must call sig_setunmask() at all the exit points of this function,
 * else all hell shall break loose for this process.
 * 
 */
int pvfsdev_enqueue(struct pvfs_upcall *in_upcall, 
						  struct pvfs_downcall *out_downcall)
{

	struct pvfsdev_upq_entry *q_upcall = NULL;
	struct pvfsdev_seq_entry *q_seq = NULL;
	struct pvfsdev_seq_entry *q_seq_probe = NULL;
	DECLARE_WAITQUEUE(my_wait, current);
	struct task_struct *tsk = current;
	pvfs_seq_t my_seq = 0;	
	int err = 0;
	int mem_xfer = 0;
	void *copybuf = NULL;
	int mem_read = 0;
	struct timeval begin = {0,0}, end = {0,0};
	int invalid_flag = 0, what_signal = 0;
	int intr = 0;
	sigset_t oldset;

	/* pvfs_map_buf and rw are only used in the PVFS_BUFFER_MAPPED case */
	void *pvfs_map_buf = NULL;
	int rw = -1;
	pvfs_callstat.total++;
	intr = (in_upcall->flags & PVFS_MOUNT_INTR) ? 1 : 0;

	/* sig_setunmask() MUST BE CALLED AT ALL EXIT POINTS IN THIS FUNCTION */
	sig_setmask(&oldset, intr); 


	/* nr_daemons tells us if there is a user-space pvfsd listening for 
	 * upcalls.  If there isn't, we bail out.
	 */
	if (nr_daemons < 1) {
		PERROR("pvfsdev: pvfsd does not appear to be running.\n");
		sig_setunmask(&oldset, intr);
		return -ENODEV;
	}
	
	q_seq = kmalloc(sizeof(struct pvfsdev_seq_entry), GFP_KERNEL);
	if (!q_seq) {
		PERROR("pvfsdev: out of memory.\n");
		sig_setunmask(&oldset, intr);
		return -ENOMEM;
	}
	q_seq->invalid = 0;

	/* here we are going to check to see if the upcall involves a
	 * memory transfer- if so we are going to go ahead and set up the
	 * pvfs io mapping, if needed.
	 */
	switch (in_upcall->type) {
	case TEST_RD_XFER:
	case READ_OP:
	case WRITE_OP:
	case TEST_WR_XFER:
		/* note that writes are a two step operation.  When we
		 * submit this upcall, the sequence structure needs to be
		 * set so that the step is zero.  
		 */
		if ((in_upcall->xfer.to_kmem != 0) || (is_fake_useraddr())) {
			/* not a user-space buffer; don't map */
			q_seq->kernelbuf.ptr = in_upcall->xfer.ptr;
			q_seq->kernelbuf.size = in_upcall->xfer.size;
			q_seq->kernel_read = 1;
			q_seq->step = 0;
			break;
		}

		/* wait on buffer resources and allocate them.  Mapping of user space
       * buffers into kernel space (in the PVFS_BUFFER_MAPPED case) is handled
       * a little further down 
		 */
		err = setup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
		if (err) {
			PERROR("pvfsdev: setup_buffer() failure.\n");
			kfree(q_seq);
			sig_setunmask(&oldset, intr);
			return err;
		}


		/* buffers are all set; let's go. */
		if ((in_upcall->type == WRITE_OP) || (in_upcall->type == TEST_WR_XFER))
		{
			/* file system write operation */
			if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
				rw = PVFS_BUF_READ; /* only used by the mapped technique */
			}
			else {
				err = copy_from_user(copybuf, in_upcall->xfer.ptr, 
											in_upcall->xfer.size);
				if (err) {
					PERROR("pvfsdev: pvfs memory transfer failure.\n");
					cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
					kfree(q_seq);
					sig_setunmask(&oldset, intr);
					return err;
				}
			}
		}
		else {
			/* file system read operation */
			if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
				rw = PVFS_BUF_WRITE;
			}
			else {
				/* this is a read xfer we need to handle later */
				mem_read = 1;
			}
		}

		if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
#ifdef HAVE_PVFS_BUFMAP
			/* map the user buffer into kernel space and point the queue entry
			 * to the right buffer
			 */
			err = pvfs_map_userbuf(rw, in_upcall->xfer.ptr,
										  in_upcall->xfer.size, &pvfs_map_buf);
			if (err) {
				PERROR("pvfsdev: pvfs_map_userbuf() failure.\n");
				cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
				kfree(q_seq);
				sig_setunmask(&oldset, intr);
				return err;
			}
			q_seq->iobuf = pvfs_map_buf;
#endif
		}
		else {
			/* point the queue entry to the right buffer */
			q_seq->iobuf = copybuf;
		}
		q_seq->step = 0;
		q_seq->kernel_read = 0;
		mem_xfer = 1;
		
		break;
	case GETDENTS_OP:
		if (in_upcall->xfer.to_kmem == 0) {
			/* not supported at this time */
			PERROR("pvfsdev: user-space address to GETDENTS?\n");
			kfree(q_seq);
			sig_setunmask(&oldset, intr);
			return -EINVAL;
		}
		q_seq->kernelbuf.ptr = in_upcall->xfer.ptr;
		q_seq->kernelbuf.size = in_upcall->xfer.size;
		q_seq->kernel_read = 1;
		
		break;
	case READLINK_OP:
		if (in_upcall->xfer.to_kmem == 0) {
			/* not supported at this time */
			PERROR("pvfsdev: user-space address to READLINK?\n");
			kfree(q_seq);
			return -EINVAL;
		}
		q_seq->kernelbuf.ptr = in_upcall->xfer.ptr;
		q_seq->kernelbuf.size = in_upcall->xfer.size;
		q_seq->kernel_read = 1;
		break;
	default:
		/* not a memory transfer operation */
		break;
	}
	

	/* creating the structure to put on the upcall queue */
	q_upcall = kmalloc(sizeof(struct pvfsdev_upq_entry), GFP_KERNEL);
	if (!q_upcall) {
		/* TODO */
		PERROR("pvfsdev: out of memory.\n");
		if (mem_xfer) cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
		kfree(q_seq);
		sig_setunmask(&oldset, intr);
		return -ENOMEM;
	}
	q_upcall->pending_up = *in_upcall;

	/* grab a sequence number atomically */
	my_seq = inc_locked_seq();
	q_upcall->pending_up.seq = my_seq;
	q_seq->seq = my_seq;

	/* atomically modify the sequence number list.  this is done to let us keep
    * sequence numbers around in case an upcall is aborted for some reason 
	 *
	 * (there might be some other reason also???)
    */
	down(&seq_sem);
	/* make sure that we don't have a duplicate of this sequence
	 * number (from an old request) still in the list
	 */
	err = seq_is_pending(&seq_pending, my_seq, &invalid_flag);
	if(err){
		if(invalid_flag) {
			PERROR("pvfsdev: stale request purged from seq_pending.\n");
			remove_seq_pending(&seq_pending, my_seq);
		}
		else {
			up(&seq_sem);
			PERROR("pvfsdev: seq_pending list overrun - bailing out!\n");
			sig_setunmask(&oldset, intr);
			return(-EINVAL);
		}
	}
	list_add(&(q_seq->seq_list), &seq_pending);
	up(&seq_sem);
	
	/* atomically add ourselves to the upcall queue */
	spin_lock(&pvfsdev_up_lock);
	list_add(&(q_upcall->req_list), &up_pending);
	spin_unlock(&pvfsdev_up_lock);

	/* wake up processes listening on pvfsdev for requests (pvfsd) */
	wake_up_interruptible(&pvfsdev_waitq);
	
	/* now I want to wait here until a downcall comes in 
	 * that matches the sequence number.  i stay on the waitqueue until 
	 * I explicitly remove myself.
	 */
	my_wait.task = tsk;
	add_wait_queue(&pvfsdev_seq_wait, &my_wait);
	/* this variable is set thru the sysctl interface */
	if(pvfs_upcall_timestamping) {
		 do_gettimeofday(&begin);
	}
seq_repeat:
	set_current_state(TASK_INTERRUPTIBLE);

	/* this lets this mechanism be interrupted if needed */
	/* but don't interrupt a hint operation!  This is necessary
	 * to correctly close sockets and such on program exit.
	 */
	/* check for bad signals.  if we got one, bail out gracefully */
	if ((check_for_bad_signals(&what_signal))) {
		PDEBUG(D_PSDEV, "Process %d inside pvfsdev (pvfsdev_enqueue): recieved signal %d- bailing out.\n", current->pid, what_signal);
		/* need to do something reasonable here... */
		/* so subsequent calls don't freak out :) */
		
		/* at least need to dequeue above stuff and make sure the
		 * module use count is correct
		 */
		remove_wait_queue(&pvfsdev_seq_wait, &my_wait);
		set_current_state(TASK_RUNNING);

		/* how far along did the pvfsd get on this operation?  Three
		 * scenarios could occur:
		 *
		 * 1) The pvfsd has not read the upcall yet.  We need to dequeue
		 * the upcall and then dequeue and destroy the sequence entry that
		 * matches it.
		 *
		 * 2) The pvfsd has read the upcall, but has not yet produced a
		 * downcall in response.  That means  that we should invalidate 
		 * the sequence number.  The pvfsdev_write function
		 * will detect that the sequence was invalidated and silently eat
		 * the matching downcall.
		 *
		 * 3) The pvfsd has read the upcall, and has already completed the
		 * downcall response. In this case, we need to get the
		 * downcall out of the queue and destroy it before we return.
		 */

		/* see if we can find the upcall and remove it */
		err = up_dequeue(&up_pending, my_seq);
		if(err == 0) {
			/* the upcall has not been read */
			/* scenario 1. */
			down(&seq_sem);
			remove_seq_pending(&seq_pending, my_seq);
			up(&seq_sem);
		}
		else {
			/* the upcall has been read */
			down(&seq_sem);
			/* has the downcall completed? */
			q_seq_probe = match_seq_entry(&seq_pending, my_seq);
			if(q_seq_probe){
				/* downcall not completed */
				/* scenario 2 */
				invalidate_seq_pending(&seq_pending, my_seq);
			}
			else{
				/* downcall has completed */
				/* scenario 3 */
				down_dequeue(&down_pending, my_seq, out_downcall); 
			}

			up(&seq_sem);
		}

		if (mem_xfer) cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
		if(pvfs_upcall_timestamping) {
			 do_gettimeofday(&end);
			 if (end.tv_usec < begin.tv_usec) {
				 end.tv_usec += 1000000; end.tv_sec--;
			 }
			 end.tv_sec  -= begin.tv_sec;
			 end.tv_usec -= begin.tv_usec;
			 /* update the upcall statistics */
			 pvfs_upcall_stats(in_upcall->type, (end.tv_sec * 1000000 + end.tv_usec));
		}
		sig_setunmask(&oldset, intr);
		return -EINTR;
	}
	
	/* here I need to check the seq_pending list and see if my sequence
	 * number has been removed, indicating a matching downcall.  if not we call
	 * schedule and go back to waiting.
	 */
	down(&seq_sem);
	err = seq_is_pending(&seq_pending, my_seq, NULL);
	up(&seq_sem);
	if (err) {
		schedule();
		goto seq_repeat;
	}
	remove_wait_queue(&pvfsdev_seq_wait, &my_wait);
	set_current_state(TASK_RUNNING);
	if(pvfs_upcall_timestamping) {
		 do_gettimeofday(&end);
		 if (end.tv_usec < begin.tv_usec) {
			 end.tv_usec += 1000000; end.tv_sec--;
		 }
		 end.tv_sec  -= begin.tv_sec;
		 end.tv_usec -= begin.tv_usec;
		 /* update the upcall statistics */
		 pvfs_upcall_stats(in_upcall->type, (end.tv_sec * 1000000 + end.tv_usec));
	 }


	/* now I need to pull my matched downcall off of the downcall
	 * list and return
	 */
	err = down_dequeue(&down_pending, my_seq, out_downcall); 
	if (err < 0) {
		if (mem_xfer) cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
		sig_setunmask(&oldset, intr);
		return err;
	}

	/* if the operation was a read, copy the data received from the daemon
	 * into the user buffer if necessary (i.e. not mapped)
	 */
   if (mem_read && (pvfs_buffer == PVFS_BUFFER_DYNAMIC 
						  || pvfs_buffer == PVFS_BUFFER_STATIC)) 
	{
		err = copy_to_user(in_upcall->xfer.ptr, copybuf, in_upcall->xfer.size);
	}

   /* if we did a transfer, clean up the buffer and wake people up */
	if (mem_xfer) cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
	sig_setunmask(&oldset, intr);
	return err;
}

/* setup_buffer()
 *
 * Used to allocate buffer space in the dynamic case, or to grab the
 * static_buf_sem semaphore in the static case.  We don't map the user buffer
 * for the mapped case yet, as there are some additional checks that must
 * first be made (SHOULD WE MOVE THE MAPPING INTO THIS FUNCTION?)
 *
 * The first half of this function handles checking resource limitations in
 * the mapped and dynamic case.  
 *
 * NOTE: pvfs_map_bufp is currently unused, as we're not doing the mapping in
 * here yet.
 */
static int setup_buffer(struct pvfs_upcall *in_upcall,
								void **copybufp, void **pvfs_map_bufp)
{
	int what_signal = 0;
	struct task_struct *tsk = current;
	DECLARE_WAITQUEUE(my_wait, current);

	/* This following bit of code makes sure that the sum of 
	 * all of the buffers used for the MAPPED or DYNAMIC buffers
	 * doesn't exceed the maxsz defined at module load time.
	 */
	if (pvfs_buffer == PVFS_BUFFER_DYNAMIC 
		 || pvfs_buffer == PVFS_BUFFER_MAPPED) 
	{
		
		spin_lock(&pvfsdev_buffer_pool_lock);
		if ((pvfsdev_buffer_size + in_upcall->xfer.size) <= pvfs_maxsz) {
			pvfsdev_buffer_size += in_upcall->xfer.size;
			spin_unlock(&pvfsdev_buffer_pool_lock);
		}
		else {
			spin_unlock(&pvfsdev_buffer_pool_lock);
			/* uh oh, we have exceeded the max buffer size available */
			PDEBUG(D_SPECIAL, "exceeded buffer size: sleeping.\n");	
			my_wait.task = tsk;
			add_wait_queue(&pvfsdev_buffer_pool_wait, &my_wait);
		dynamic_repeat:
			set_current_state(TASK_INTERRUPTIBLE);
			
			/* check for bad signals and bail out if we got one */
			if ((check_for_bad_signals(&what_signal))) {
				PDEBUG(D_PSDEV, "Process %d inside pvfsdev (setup_buffer): recieved signal %d- bailing out.\n", current->pid, what_signal);
				remove_wait_queue(&pvfsdev_buffer_pool_wait, &my_wait);
				set_current_state(TASK_RUNNING);
				return -EINTR; 
			}
				
			/* now we can look again to see if we can allocate a buffer */
			spin_lock(&pvfsdev_buffer_pool_lock);
			if ((pvfsdev_buffer_size + in_upcall->xfer.size) > pvfs_maxsz) {
				spin_unlock(&pvfsdev_buffer_pool_lock);
				schedule();
				/* go back to sleep if we can't */
				goto dynamic_repeat;
			}
			
			/* the memory is available- let's rock! */
			pvfsdev_buffer_size += in_upcall->xfer.size;
			spin_unlock(&pvfsdev_buffer_pool_lock);
			
			remove_wait_queue(&pvfsdev_buffer_pool_wait, &my_wait);
			set_current_state(TASK_RUNNING);
		}
	}
	
	/* Here we allocate our buffer for the dynamic copy, or simply
	 * set the pointer to point to the static buffer in the static
	 * case.  This value isn't used by the mapped buffer code.
	 */
	if (pvfs_buffer == PVFS_BUFFER_DYNAMIC) {
		*copybufp = (void*)vmalloc(in_upcall->xfer.size);

		if (*copybufp == NULL) {
			PERROR("pvfsdev: could not allocate memory!\n");
			spin_lock(&pvfsdev_buffer_pool_lock);
			pvfsdev_buffer_size -= in_upcall->xfer.size;
			spin_unlock(&pvfsdev_buffer_pool_lock);
			return -ENOMEM;
		}
	}
	else if (pvfs_buffer == PVFS_BUFFER_STATIC) {
		/* go ahead and decrement the semaphore while we are at it */
		down(&static_buf_sem);
		*copybufp = static_buf;
	}

	/* this would be the place to map the user buffer if we wanted to
	 * do it in this function.
	 */

	return 0;
}


/* cleanup_buffer()
 *
 * Used to deallocate buffer space in the dynamic case, or to unmap user
 * buffers in the mapped case, or to release the static_buf_sem semaphore in
 * the static case.
 *
 * This function also handles waking up other processes waiting on these
 * resources.  Both the mapped and dynamic schemes use the
 * pvfsdev_buffer_pool_wait wait queue to wait on resources to be freed, so we
 * wake up the guys on that queue in here.  
 */
static inline void cleanup_buffer(struct pvfs_upcall *in_upcall,
											 void **copybufp, void **pvfs_map_bufp)
{
	if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
#ifdef HAVE_PVFS_BUFMAP
		if (*pvfs_map_bufp) {
			pvfs_unmap_userbuf(*pvfs_map_bufp);
			*pvfs_map_bufp = NULL;
		}
		spin_lock(&pvfsdev_buffer_pool_lock);
		pvfsdev_buffer_size -= in_upcall->xfer.size;
		spin_unlock(&pvfsdev_buffer_pool_lock);
		wake_up_interruptible(&pvfsdev_buffer_pool_wait);
#endif
	}
	else if (pvfs_buffer == PVFS_BUFFER_DYNAMIC) {
		if (*copybufp) {
			vfree(*copybufp);
			*copybufp = NULL;
		}
		spin_lock(&pvfsdev_buffer_pool_lock);
		pvfsdev_buffer_size -= in_upcall->xfer.size;
		spin_unlock(&pvfsdev_buffer_pool_lock);
		wake_up_interruptible(&pvfsdev_buffer_pool_wait);
	}
	else if (pvfs_buffer == PVFS_BUFFER_STATIC) {
		up(&static_buf_sem);
	}
	return;
}

/* check_for_bad_signals()
 *
 * This function checks for pending signals after a process has woken up.
 *
 * Returns 0 if everything is ok, -EINTR otherwise and the signal number in sig_nr.
 */
static inline int check_for_bad_signals(int *sig_nr)
{
	int i;
	if (signal_pending(current)) {
		/* 
		 *  This is really ugly. I dont understand what to substitute instead of 32.
		 *  In fact 32 is correct only on x86 and ia64 as far as user-space is concerned
		 */
		for(i = 1; i <= 32; i++) {
			if(sigismember(&(current->pending.signal), i)) {
				*sig_nr = i;
				return -EINTR;
			}
		}
		/* give up if you cannot find out */
		*sig_nr = -1;
		/* Is this possible at all?? */
		PERROR("check_signal says signal_pending, but could not find out which signal!!\n");
		return -EINTR; 
	}
	return 0;
}


/* pvfsdev_enqueue_noresponse()
 *
 * This function is used to queue an upcall for which there will be no
 * matching downcall.  This is indicated to the daemon by a -1 sequence
 * number, which is stored in the queue structure (q_upcall). 
 */
int pvfsdev_enqueue_noresponse(struct pvfs_upcall *in_upcall)
{	
	struct pvfsdev_upq_entry *q_upcall;

   /* nr_daemons tells us if there is a user-space pvfsd listening for 
	 * upcalls.  If there isn't, we bail out.
	 */
	if (nr_daemons < 1) {
		return -ENODEV;
	}

	q_upcall = kmalloc(sizeof(struct pvfsdev_upq_entry), GFP_KERNEL);
	if(!q_upcall){
		return -ENOMEM;
	}

	/* this q_upcall structure is the element that sits in the queue */
	q_upcall->pending_up = *in_upcall;
	q_upcall->pending_up.seq = -1;
	
	/* add this upcall to the queue */
	spin_lock(&pvfsdev_up_lock);
	list_add(&(q_upcall->req_list), &up_pending);
	spin_unlock(&pvfsdev_up_lock);
	
	/* wake up processes reading or selecting on our device */
	wake_up_interruptible(&pvfsdev_waitq);
	
	return 0;
}


/* seq_is_pending()
 *
 * use this to find out if a sequence number is in the seq list
 *
 * now the invalid integer will be set to alert the caller that they
 * have matched a sequence number that has been declared invalid.
 * if they caller does not want to know about that they should set
 * that pointer to NULL
 *
 * IMPORTANT: this function assumes that the caller holds a semaphore
 * protecting the sequence list!
 */
static int seq_is_pending(struct list_head *my_list, pvfs_seq_t my_seq,
								  int *invalid)
{
	struct list_head *check;
	struct pvfsdev_seq_entry *entry;
	
	for (check = my_list->next; check != my_list; check = check->next)
   {
		entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);

		if (entry->seq == my_seq) {
			if (invalid != NULL) {
				*invalid = entry->invalid;
			}
			return 1;
		}
	}
	return 0;
}


/* down_dequeue()
 *
 * use this to figure out which downcall to dequeue based on which
 * one matches your sequence number-return null on error...
 *
 * now also returns the fs from the downcall side :)
 *
 * Returns 0 on success, -errno on failure.
 */
static int down_dequeue(struct list_head *my_list, pvfs_seq_t my_seq,
								struct pvfs_downcall *out_downcall)
{ 
	struct list_head *check;
	struct pvfsdev_downq_entry *entry;

	spin_lock(&pvfsdev_down_lock);

	/* iterate through list of entries looking for sequence number;
	 * we expect it to be near the head of the list 
    */
	for (check = my_list->next; check != my_list; check = check->next) 
	{
		entry = list_entry(check, struct pvfsdev_downq_entry, req_list);

		if (entry->pending_down.seq == my_seq) {
			/* found it.  return copy of downcall, remove entry from list, 
			 * free downcall entry structure, unlock list, and return.
			 */
			*out_downcall = entry->pending_down;
			list_del(check);
			kfree(entry);
			/* we don't have to kfree() pending_down because it is a part of the
			 * pvfsdev_downq_entry structure
			 */
			spin_unlock(&pvfsdev_down_lock);
			return 0;
		}
	}

	/* looked through the whole list and didn't find a match */
	spin_unlock(&pvfsdev_down_lock);

	PERROR("pvfsdev: didn't find matching downcall in queue!\n");
	return -EINVAL;
}
		

/* pvfsdev_down_enqueue()
 *
 * this adds downcalls to the incoming queue
 *
 * IMPORTANT!  This function assumes that the caller holds a semaphore
 * protecting the sequence list before it is called!
 *
 * Returns 0 on success, -errno on failure.
 */
static int pvfsdev_down_enqueue(struct pvfs_downcall *in_downcall)
{
	struct pvfsdev_downq_entry *q_downcall;
	int ret = -1;

	/* we also need to handle removing the sequence number from the
	 * list of pending sequence numbers for upcall/downcall pairs
	 */

	q_downcall = kmalloc(sizeof(struct pvfsdev_downq_entry), GFP_KERNEL);
	if(!q_downcall){
		return -ENOMEM;
	}
	
	/* make a copy of the downcall */
	q_downcall->pending_down = *in_downcall;

	spin_lock(&pvfsdev_down_lock);
	list_add(&(q_downcall->req_list), &down_pending);
	spin_unlock(&pvfsdev_down_lock);

	/* take the sequence number out of the sequence number list */
	ret = remove_seq_pending(&seq_pending, in_downcall->seq);
	if(ret < 0){
		PDEBUG(D_DOWNCALL, "pvfsdev: remove of seq_entry failed... \n");
	}

	/* wake stuff up - see poll function */
	wake_up_interruptible(&pvfsdev_seq_wait);
	
	return 0;
}


/* remove_seq_pending()
 *
 * use this to remove a sequence number from the pending sequence
 * list
 *
 * IMPORTANT!  This function assumes that the caller already holds a
 * semaphore to protect the sequence list!
 *
 * Returns 0 on success, -errno on failure.
 */
static int remove_seq_pending(struct list_head *my_list, pvfs_seq_t my_seq)
{
	struct list_head *check;
	struct pvfsdev_seq_entry *entry;

	for (check = my_list->next; check != my_list; check = check->next) 
	{
		entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);

		if (entry->seq == my_seq) {
			list_del(check);
			kfree(entry);
			return 0;
		}
	}

	return -EINVAL;
}


/* inc_locked_seq()
 *
 * this simply increments the communication sequence number and
 * returns its new value.
 */
static pvfs_seq_t inc_locked_seq()
{
	pvfs_seq_t ret = 0;

	spin_lock(&pvfsdev_current_seq_lock);
	if (current_seq < PVFSDEV_MAX_SEQ) {
		current_seq++;
	}
	else {
		current_seq = 1;
	}
	ret = current_seq;
	spin_unlock(&pvfsdev_current_seq_lock);

	return ret;
}


/* up_dequeue()
 *
 * this function dequeues specific pending upcalls- mostly used for
 * cleanup mechanisms
 *
 * Returns 0 on success, -errno on failure.
 */
static int up_dequeue(struct list_head *my_list, pvfs_seq_t my_seq)
{ 
	struct list_head *check;
	struct pvfsdev_upq_entry *entry;
	
	spin_lock(&pvfsdev_up_lock);

	/* iterate through list of entries looking for matching sequence number.
	 * we go through the list from oldest to newest, because we expect we're 
	 * looking for an old entry.
	 */
	for (check = up_pending.prev; check != &up_pending; check = check->prev) 
	{
		entry = list_entry(up_pending.prev, struct pvfsdev_upq_entry, req_list);

		if (entry->pending_up.seq == my_seq) {
			list_del(up_pending.prev);
			kfree(entry);
			spin_unlock(&pvfsdev_up_lock);
			return 0;
		}
	}

	spin_unlock(&pvfsdev_up_lock);
	return -EINVAL;
}


/* match_seq_entry()
 *
 * returns a pointer to the seq_entry with a matching sequence number,
 * or NULL if not found in the list.
 *
 * IMPORTANT! This function assumes that the caller holds a sequence
 * list semaphore when calling this function.
 */
static struct pvfsdev_seq_entry *match_seq_entry(struct list_head *my_list, 
																 pvfs_seq_t my_seq)
{

	struct list_head *check;
	struct pvfsdev_seq_entry *entry;

	for (check = my_list->next; check != my_list; check = check->next) 
	{
		entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);

		if (entry->seq == my_seq) {
			return entry;
		}
	}

	return NULL;
}

/* invalidate_seq_pending()
 *
 * this function will be used to set the invalid flag in a sequence
 * list entry. future operations on that sequence will be
 * aware that a previous step in the transfer has failed.
 *
 * IMPORTANT!  This function assumes that the caller is holding a
 * semaphore to protect the sequence list.
 *
 * Returns 0 on success, -errno on failure.
 */
static int invalidate_seq_pending(struct list_head *my_list, pvfs_seq_t my_seq)
{

	struct list_head *check;
	struct pvfsdev_seq_entry *entry;

	for (check = my_list->next; check != my_list; check = check->next) 
	{
		entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);

		if (entry->seq == my_seq) {
			entry->invalid = 1;
			return 0;
		}
	}

	return -EINVAL;
}

/* is_fake_useraddr()
 *
 * This function tests for the case where someone has used the set_fs()
 * trick to pass a kernel address to a function that normally takes a
 * user address.
 *
 * Returns 0 if this appears to be a real user address, non-zero if not.
 */
static int is_fake_useraddr(void)
{
	mm_segment_t check_fs;
	mm_segment_t check_ds;

	check_fs = get_fs();
	check_ds = get_ds();
	
	if (check_fs.seg == check_ds.seg) {
		PDEBUG(D_SPECIAL, "is_fake_useraddr detected fake address\n");
		return 1;
	}
	return 0;
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
