/*
 * copyright (c) 1999 Rob Ross and Phil Carns, all rights reserved.
 *
 * Written by Rob Ross and Phil Carns, funded by Scyld Computing.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.clemson.edu
 *           Phil Carns  pcarns@parl.clemson.edu
 */

#define __NO_VERSION__

#include "config.h"
#include "pvfs_kernel_config.h"
#include "pvfs_linux.h"
#include "ll_pvfs.h"
#include "pvfs_proc.h"

#ifdef HAVE_LINUX_LOCKS_H
#include <linux/locks.h>
#endif
#ifdef HAVE_LINUX_SMP_LOCK_H
#include <linux/smp_lock.h>
#endif

/* dir inode-ops */
static int pvfs_readlink(struct dentry*, char*, int);
static int pvfs_followlink(struct dentry*, struct nameidata *);

struct inode_operations pvfs_symlink_inode_operations = {
	readlink: pvfs_readlink, /* read link */
	follow_link: pvfs_followlink, /* follow link */
	revalidate: pvfs_revalidate_inode, /* revalidate inode */
};

static int pvfs_symlink_filler(struct inode *inode, struct page *page)
{
	void *buffer = kmap(page);
  	int error;
	/* Make sure that it is NUL terminated */
  	error = ll_pvfs_readlink(pvfs_inop(inode), buffer, PAGE_CACHE_SIZE);
	if (error < 0)
   	goto error;
  	SetPageUptodate(page);
  	kunmap(page);
  	UnlockPage(page);
  	return 0;

error:
  	SetPageError(page);
  	kunmap(page);
  	UnlockPage(page);
  	return -EIO;
}

static char *pvfs_getlink(struct inode *inode, struct page **ppage)
{
	struct page *page;
  	u32 *p;

  	/* Caller revalidated the directory inode already. */
  	page = read_cache_page(&inode->i_data, 0, (filler_t *)pvfs_symlink_filler, inode);
	if (IS_ERR(page))
  	 	goto read_failed;
 	if (!Page_Uptodate(page))
    	goto getlink_read_error;
  	*ppage = page;
  	p = kmap(page);
  	return (char*)p;
  
getlink_read_error:
	page_cache_release(page);
  	return ERR_PTR(-EIO);
read_failed:
  	return (char*)page;
}

static int pvfs_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	struct inode *inode = dentry->d_inode;
 	struct page *page = NULL;
  	int res;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.readlink++;       
	res = vfs_readlink(dentry, buffer, buflen, pvfs_getlink(inode,&page));
  	if (page) {
    	kunmap(page);
    	page_cache_release(page);
  	}
  	return res;
}

static int pvfs_followlink(struct dentry *dentry, struct nameidata *nd)
{
	struct inode *inode = dentry->d_inode;
	struct page *page = NULL;
  	int res;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.readlink++;       
	res = vfs_follow_link(nd, pvfs_getlink(inode, &page));
  	if (page) {
   	kunmap(page);
	  	page_cache_release(page);
  	}
  	return res;
}
